==> spice-js-1.1.17.tgz: package/src/models/SpiceModel.js <==
"use strict";
var SDate = require('sonover-date'),
	UUID = require('uuid'),
	_ = require('lodash'),
	_database = Symbol(),
	_props = Symbol();
import Database from "spice-couchbase";
export default class SpiceModel {

	constructor(args = {}) {
		try {
			this.type = "";
			this.deleted = false;
			if (args.args) {
				delete args.args['ctx'];
			}
			this[_database] = new Database(args.connection || 'default', require('couchbase').N1qlQuery, require('couchbase').SearchQuery);
			this[_props] = args.props;
			for (let i in args.props) {
				if (args.args != undefined) {
					if (args.args[i] != undefined) {
						switch (args.props[i].type) {
							case 'string': {
								this[i] = args.args[i];
								break;
							}
							case 'number': {
								this[i] = _.isNumber(args.args[i]) ? args.args[i] : Number(args.args[i]);
								break;
							}
							case 'boolean': {
								console.log("Boolean In Spice", args.args[i])
								this[i] = _.isBoolean(args.args[i]) ? args.args[i] : args.args[i] == 'true';
								break;
							}
							case 'date': {
								this[i] = args.args[i];
								break;
							}
							case 'array': {
								this[i] = _.isArray(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
								break;
							}
							case 'object': {
								//console.log(args.args[i])
								this[i] = _.isObject(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
								break;
							}
							default: {
								this[i] = args.args[i];
							}
						}
					}
				}
			}
			console.log("Constructure log", args.args);
		} catch (e) {
			console.log(e.stack);
		}

	}

	get database() {
		return this[_database];
	}

	set database(value) {
		this[_database] = value;
	}

	get props() {
		return this[_props];
	}

	set props(value) {
		this[_props] = value;
	}

	is_date(item) {
		return item.length > 27 && item.indexOf("|") > -1
	}

	makeQueryFromFilter(filter) {
		let return_string = "";
		for (let key in filter) {
			let item_string = "";
			if (filter[key].length > 1) {
				item_string = "(";
			}

			let item_count = 0;

			for (let item of filter[key]) {

				if (item_count > 0) {
					item_string = item_string + " OR "
				}

				if (this.is_date(item)) {
					let to_index = item.indexOf("|");
					let first = item.substring(0, to_index);
					let second = item.substring(to_index + 1, item.length);
					item_string = item_string + " " + key + " BETWEEN '" + first + "' AND '" + second + "'";
				} else {
					item_string = item_string + " " + key + " LIKE '" + item + "'";
				}

				item_count++;
			}

			if (filter[key].length > 1) {
				item_string = item_string + ")";
			}
			if (return_string != "") {
				return_string = return_string + " AND ";
			}
			return_string = return_string + item_string;
		}
		return return_string;
	}

	async get(args) {
		try {
			if (!args) {
				args = {};
			}
			if (_.isString(args.id)) {
				await this.run_hook(args, 'get', 'before');
				//console.log("getting ", args, "From", this.database.bucket);
				let results = await this.database.get(args.id);
				if (results.type != this.type) {
					throw new Error(`${this.type} does not exist`);
				}
				if (results.deleted == undefined || results.deleted == false) {
					await this.run_hook(results, 'get', 'after');
					await this.do_serialize(results, 'read');
					return results;
				} else {
					throw new Error(`${this.type} does not exist`);
				}
			}
		} catch (e) {
			throw e;
		}
	}

	async query(query) {
		try {
			let results = await this.database.query(query);
			return results
		} catch (error) {
			throw error;
		}
	}



	async getMulti(args) {
		try {
			if (!args) {
				args = {};
			}

			await this.run_hook(this, 'get', 'before');
			_.remove(args.ids, o => {
				return o == undefined;
			})
			let results = []
			if (args.ids.length > 0) {
				//console.log("Here", args.ids)
				results = await this.database.multi_get(args.ids, true);
			}
			let removed = _.remove(results, o => {
				return o.type != this.type;
			})
			if (_.isArray(results)) {
				for (let i of results) {
					await this.run_hook(i, 'get', 'after', args.context);
					await this.do_serialize(i, 'read');
				}
			} else {
				let i = results;
				await this.run_hook(i, 'get', 'after', args.context);
				await this.do_serialize(i, 'read');
			}

			return results;

		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}

	async exist(data) {
		try {
			if (_.isString(data)) {
				let result = await this.database.get(data);
				if (result.type != this.type) {
					return false;
				}
			} else {
				if (data.type != this.type) {
					return false;
				}
			}
			return true;
		} catch (e) {
			throw e;
		}
	}

	async update(args) {
		try {
			this.updated_at = new SDate().now();
			//console.log("Id in update", args.id, this)
			let results = await this.database.get(args.id);
			//console.log("Id Results:", results)
			let item_exist = await this.exist(results);
			if (!item_exist) {
				throw new Error(`${this.type} does not exist.`);
			}
			delete results['id'];
			_.defaults(this, results);
			let cover_obj = {
				old: results,
				new: this,
				id: args.id
			}
			if (args.skip_hooks != true) {
				await this.run_hook(cover_obj, 'update', 'before');
			}

			//console.log("after hook")
			if (args.skip_write_serialize != true && args.skip_serialize != true) {
				await this.do_serialize(this, 'write', cover_obj.new);
			}
			//console.log("After serialize")
			await this.database.update(args.id, this);
			//console.log("After update")
			if (args.skip_hooks != true) {
				await this.run_hook({
					...this,
					id: args.id
				}, 'update', 'after');
			}
			if (args.skip_read_serialize != true && args.skip_serialize != true) {
				await this.do_serialize(this, 'read');
			}
			this.id = args.id;
			return this;
		} catch (e) {
			console.log("Error on update", e, e.stack);
			throw e;
		}
	}

	async create(args) {
		try {
			let form;
			this.created_at = new SDate().now();
			if (args.body) {
				form = _.defaults({}, args.body);
				form.created_at = this.created_at;
				form.updated_at = this.created_at;
				delete form['bucket'];
			}
			this.updated_at = new SDate().now();


			let id = "";
			if (args && args.id) {
				id = args.id;
			} else {
				id = `${args.id_prefix}-${UUID.v4()}`;
			}

			await this.run_hook(form || this, 'create', 'before');
			await this.do_serialize(form || this, 'write');

			let results = await this.database.insert(id, form || this, args.expiry);
			await this.run_hook({
				...results,
				id
			}, 'create', 'after');
			await this.do_serialize(results, 'read');
			return results;
		} catch (e) {
			console.log(e.stack);
			throw e
		}

	}

	async delete(args) {
		let item_exist = await this.exist(args.id);

		if (!item_exist) {
			throw new Error(`${this.type} does not exist.`);
		}
		try {
			await this.run_hook(args, 'delete', 'before');
			let delete_response = {};
			if (args.hard) {
				delete_response = await this.database.delete(args.id);
			} else {
				let results = await this.database
					.get(args.id);
				delete results['id'];
				results.deleted = true;
				delete_response = await this.database.update(args.id, "");
			}
			await this.run_hook(args, 'delete', 'after');
			//await this.do_serialize(args, 'read');
			return {};
		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}

	toString() {
		JSON.stringify(this);
	}

	async list(args) {
		//console.log("1")
		try {
			if (!args) {
				args = {};
			}

			let query = "";

			if ((args.is_full_text && args.is_full_text === "true") || (args.is_custom_query && args.is_custom_query === "true")) {
				query = args.query;
			} else {
				if (args.filters) {
					query = await this.makeQueryFromFilter(args.filters);
				} else {
					if (args.query) {
						query = args.query + " AND (deleted = false OR deleted IS MISSING) ";
					} else {
						query = "(deleted = false OR deleted IS MISSING) ";
					}
				}
			}

			if (args.limit) {
				args.limit = Number(args.limit)
			}
			if (args.offset) {
				args.offset = Number(args.offset)
			}

			if (!args.sort) {
				args.sort = "";
			}

			await this.run_hook(this, 'get', 'before');
			let results
			if (args.is_custom_query && args.is_custom_query === "true") {
				//console.log("custom",query)
				results = await this.database.query(query);
				await this.run_hook(results, 'get', 'after');
				await this.do_serialize(results, 'read');
				//console.log("1x")
				return results;
			} else {
				if (args.is_full_text && args.is_full_text === "true") {
					results = await this.database.full_text_search(this.type, query || "", args.limit, args.offset);
				} else {
					try {
						results = await this.database.search(this.type, args.columns || "", query || "", args.limit, args.offset, args.sort, args.do_count, args.statement_consistent);
					} catch (e) {
						console.log(e.stack)
					}
				}
				try {
					await this.run_hook(results.data, 'get', 'after');
					await this.do_serialize(results.data, 'read');
				} catch (e) {
					console.log(e);
				}
				//console.log("1x")
				return results.data;
			}


		} catch (e) {
			console.log(e.stack);
			throw e;
		}

	}

	async run_hook(data, op, when, old_data) {
		try {
			if (this.hooks && this.hooks()[op] && this.hooks()[op][when]) {
				if (_.isArray(this.hooks()[op][when])) {
					for (var i of this.hooks()[op][when]) {
						data = await i(data, old_data);
					}
				} else {
					data = await this.hooks()[op][when](data, old_data);
				}
			}
			return data;
		} catch (e) {
			throw e;
		}
	}

	/*async run_hook(data, obj, context){
		try{


			if(obj) {
				if (_.isArray(obj)) {
					for (var i of obj) {
						await i.apply(this, [data, context]);
					}
				} else if (_.isFunction(obj)) {
					await obj.apply(this, [data, context]);
				}
			}
		}catch (e){
			console.log(e.stack);
		}
	}*/

	defaultCleaners() {
		return [
			async (data) => {
				try {
					if (_.isArray(data)) {
						for (let i of data) {
							delete i['deleted'];
							delete i['type'];
						}
					} else {
						delete data['deleted'];
						delete data['type'];
					}
					return data;
				} catch (e) {
					throw e;
				}
			}
		]
	}

	defaultModifiers() {
		return [
			async (data) => {
				try {
					return data;
				} catch (e) {
					throw e;
				}
			}
		]
	}

	serializer() {
		return {
			read: {
				modifiers: [
					/*async (data)=>{
						try{
							if(_.isArray(data)){
								//modifiers for arrays

							}else{

								//modifiers for single
							}
							return data;
						}catch(e){
							throw new Error(e);
						}
					}*/
				],
				cleaners: [
					//this is the default cleaner is none is specifier. add cleaner to override the default
					/*async (data)=>{
						try{
							if(_.isArray(data)){
								for(let i of data){
									delete i['deleted'];
									delete i['type'];
								}
							}else{
								delete data['deleted'];
								delete data['type'];
							}
							return data;
						}catch(e){
							throw new Error(e);
						}
					}*/
				]
			},
			write: {
				modifiers: [
					async (data) => {
						try {
							return data;
						} catch (e) {
							throw e;
						}
					}
				],
				cleaners: [
					async (data) => {
						try {
							return data;
						} catch (e) {
							throw e;
						}
					}
				]
			}
		}
	}

	hooks() {
		return {
			create: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}

				}
			},
			get: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			},
			update: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			},
			delete: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			}
		}
	}

	async do_serialize(data, type, old_data) {
		try {
			let serializer_obj = this.serializer();
			let active_modifiers = serializer_obj[type]['modifiers']

			if (!active_modifiers) {
				await this.defaultModifiers();
			}

			if (active_modifiers) {
				if (_.isArray(active_modifiers)) {
					for (let i of active_modifiers) {
						try {
							data = await i(data, old_data);
						} catch (e) {
							console.log(e.stack);
						}
					}
				} else {
					data = await active_modifiers(data, old_data);
				}
			}
			//apply cleaners
			let active_cleaners = serializer_obj[type]['cleaners']
			if (type == 'read') {
				if (!active_cleaners) {
					active_cleaners = await this.defaultCleaners();
				} else {
					if (_.isArray(active_cleaners)) {
						if (active_cleaners.length == 0) {
							active_cleaners = await this.defaultCleaners();
						}
					}
				}
			}

			if (active_cleaners) {
				if (_.isArray(active_cleaners)) {
					for (let i of active_cleaners) {
						data = await i(data);
					}
				} else {
					data = await active_cleaners(data);
				}
			}

			//get defaults from peroperties
			let defaults = {}
			if (this.props) {
				for (let i in this.props) {
					if (this.props[i].defaults != undefined && this.props[i].defaults[type] != undefined) {
						defaults[i] = this.props[i].defaults[type];
					}
				}
			}

			//apply defaults
			if (_.isArray(data)) {
				for (let i of data) {
					_.defaults(i, defaults);
				}
			} else {
				_.defaults(data, defaults);
			}

			return data;
		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}
}

==> spice-js-2.0.9.tgz: package/src/models/SpiceModel.js <==
"use strict";
var regeneratorRuntime = require('regenerator-runtime');
var SDate = require('sonover-date'),
	UUID = require('uuid'),
	_ = require('lodash'),
	_database = Symbol(),
	_props = Symbol();
export default class SpiceModel {

	constructor(args = {}) {

		try {
			var dbtype = spice.config.database.connections[args.connection].type || "couchbase";
			let Database = require(`spice-${dbtype}`);
			this.type = "";
			this.deleted = false;
			if (args.args) {
				delete args.args['ctx'];
			}
			this[_database] = new Database(args.connection || 'default');
			this[_props] = args.props;
			for (let i in args.props) {
				if (args.args != undefined) {
					if (args.args[i] != undefined) {
						switch (args.props[i].type) {
							case 'string': {
								this[i] = args.args[i];
								break;
							}
							case 'number': {
								this[i] = _.isNumber(args.args[i]) ? args.args[i] : Number(args.args[i]);
								break;
							}
							case 'boolean': {
								this[i] = _.isBoolean(args.args[i]) ? args.args[i] : args.args[i] == 'true';
								break;
							}
							case 'date': {
								this[i] = args.args[i];
								break;
							}
							case 'array': {
								this[i] = _.isArray(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
								break;
							}
							case 'object': {
								//console.log(args.args[i])
								this[i] = _.isObject(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
								break;
							}
							default: {
								this[i] = args.args[i];
							}
						}
					}
				}
			}
		} catch (e) {
			console.log(e.stack);
		}
		//}
	}

	get database() {
		return this[_database];
	}
	set database(value) {
		this[_database] = value;
	}

	get props() {
		return this[_props];
	}

	set props(value) {
		this[_props] = value;
	}

	is_date(item) {
		return item.length > 27 && item.indexOf("|") > -1
	}

	makeQueryFromFilter(filter) {
		let return_string = "";
		for (let key in filter) {
			let item_string = "";
			if (filter[key].length > 1) {
				item_string = "(";
			}

			let item_count = 0;

			for (let item of filter[key]) {

				if (item_count > 0) {
					item_string = item_string + " OR "
				}

				if (this.is_date(item)) {
					let to_index = item.indexOf("|");
					let first = item.substring(0, to_index);
					let second = item.substring(to_index + 1, item.length);
					item_string = item_string + " " + key + " BETWEEN '" + first + "' AND '" + second + "'";
				} else {
					item_string = item_string + " " + key + " LIKE '" + item + "'";
				}

				item_count++;
			}

			if (filter[key].length > 1) {
				item_string = item_string + ")";
			}
			if (return_string != "") {
				return_string = return_string + " AND ";
			}
			return_string = return_string + item_string;
		}
		return return_string;
	}

	async get(args) {
		try {
			if (!args) {
				args = {};
			}
			if (_.isString(args.id)) {
				await this.run_hook(args, 'get', 'before');
				//console.log("getting ", args, "From", this.database.bucket);
				let results = await this.database.get(args.id);

				console.log(results);
				if (results.type != this.type) {
					throw new Error(`${this.type} does not exist`);
				}
				if (results.deleted == undefined || results.deleted == false) {
					await this.run_hook(results, 'get', 'after');
					await this.do_serialize(results, 'read', {}, args);
					return results;
				} else {
					throw new Error(`${this.type} does not exist`);
				}
			}
		} catch (e) {
			throw e;
		}
	}

	async query(query) {
		try {
			let results = await this.database.query(query);
			return results
		} catch (error) {
			throw error;
		}
	}



	async getMulti(args) {
		try {
			if (!args) {
				args = {};
			}

			await this.run_hook(this, 'get', 'before');
			_.remove(args.ids, o => {
				return o == undefined;
			})
			let results = []
			if (args.ids.length > 0) {
				results = await this.database.multi_get(args.ids, true);

			}
			let removed = _.remove(results, o => {
				return o.type != this.type;
			})
			if (_.isArray(results)) {
				for (let i of results) {
					await this.run_hook(i, 'get', 'after', args.context);
					await this.do_serialize(i, 'read', {}, args);
				}
			} else {
				let i = results;
				await this.run_hook(i, 'get', 'after', args.context);
				await this.do_serialize(i, 'read', {}, args);
			}
			console.log(results);
			return results;

		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}

	async exist(data) {
		try {
			if (_.isString(data)) {
				let result = await this.database.get(data);
				if (result.type != this.type) {
					return false;
				}
			} else {
				if (data.type != this.type) {
					return false;
				}
			}
			return true;
		} catch (e) {
			throw e;
		}
	}

	async update(args) {
		try {
			this.updated_at = new SDate().now();
			//console.log("Id in update", args.id, this)
			let results = await this.database.get(args.id);
			//console.log("Id Results:", results)
			let item_exist = await this.exist(results);
			if (!item_exist) {
				throw new Error(`${this.type} does not exist.`);
			}
			delete results['id'];
			_.defaults(this, results);
			let cover_obj = {
				old: results,
				new: this,
				id: args.id
			}
			if (args.skip_hooks != true) {
				await this.run_hook(cover_obj, 'update', 'before');
			}

			//console.log("after hook")
			if (args.skip_write_serialize != true && args.skip_serialize != true) {
				await this.do_serialize(this, 'write', cover_obj.new, args);
			}
			//console.log("After serialize")
			await this.database.update(args.id, this);
			//console.log("After update")
			if (args.skip_hooks != true) {
				await this.run_hook({
					...this,
					id: args.id
				}, 'update', 'after');
			}
			if (args.skip_read_serialize != true && args.skip_serialize != true) {
				await this.do_serialize(this, 'read', {}, args);
			}
			this.id = args.id;
			return this;
		} catch (e) {
			console.log("Error on update", e, e.stack);
			throw e;
		}
	}

	async create(args) {
		try {
			let form;
			this.created_at = new SDate().now();
			if (args.body) {
				form = _.defaults({}, args.body);
				form.created_at = this.created_at;
				form.updated_at = this.created_at;
				delete form['bucket'];
			}
			this.updated_at = new SDate().now();


			let id = "";
			if (args && args.id) {
				id = args.id;
			} else {
				id = `${args.id_prefix}-${UUID.v4()}`;
			}

			await this.run_hook(form || this, 'create', 'before');
			await this.do_serialize(form || this, 'write', {}, args);

			let results = await this.database.insert(id, form || this, args.expiry);
			await this.run_hook({
				...results,
				id
			}, 'create', 'after');
			await this.do_serialize(results, 'read', {}, args);
			return results;
		} catch (e) {
			console.log(e.stack);
			throw e
		}

	}

	async delete(args) {
		let item_exist = await this.exist(args.id);

		if (!item_exist) {
			throw new Error(`${this.type} does not exist.`);
		}
		try {
			await this.run_hook(args, 'delete', 'before');
			let delete_response = {};
			if (args.hard) {
				delete_response = await this.database.delete(args.id);
			} else {
				let results = await this.database
					.get(args.id);
				delete results['id'];
				results.deleted = true;
				delete_response = await this.database.update(args.id, "");
			}
			await this.run_hook(args, 'delete', 'after');
			//await this.do_serialize(args, 'read', {}, args);
			return {};
		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}

	toString() {
		JSON.stringify(this);
	}

	async list(args) {
		try {
			if (!args) {
				args = {};
			}

			let query = "";

			if ((args.is_full_text && args.is_full_text === "true") || (args.is_custom_query && args.is_custom_query === "true")) {
				query = args.query;
			} else {
				if (args.filters) {
					query = await this.makeQueryFromFilter(args.filters);
				} else {
					if (args.query) {
						query = args.query + " AND (deleted = false OR deleted IS MISSING) ";
					} else {
						query = "(deleted = false OR deleted IS MISSING) ";
					}
				}
			}

			if (args.limit) {
				args.limit = Number(args.limit)
			}
			if (args.offset) {
				args.offset = Number(args.offset)
			}

			if (!args.sort) {
				args.sort = "";
			}

			await this.run_hook(this, 'get', 'before');
			let results
			if (args.is_custom_query && args.is_custom_query === "true") {
				//console.log("custom",query)
				results = await this.database.query(query);
				await this.run_hook(results, 'get', 'after');
				await this.do_serialize(results, 'read', {}, args);
				//console.log("1x")
				return results;
			} else {
				if (args.is_full_text && args.is_full_text === "true") {
					results = await this.database.full_text_search(this.type, query || "", args.limit, args.offset);
				} else {
					try {
						results = await this.database.search(this.type, args.columns || "", query || "", args.limit, args.offset, args.sort, args.do_count, args.statement_consistent);
					} catch (e) {
						console.log(e.stack)
					}
				}
				try {
					await this.run_hook(results.data, 'get', 'after');
					await this.do_serialize(results.data, 'read', {}, args);
				} catch (e) {
					console.log(e);
				}
				//console.log("1x")
				return results.data;
			}


		} catch (e) {
			console.log(e.stack);
			throw e;
		}

	}



	async list(args) {
		try {
			if (!args) {
				args = {};
			}

			let query = "";

			if ((args.is_full_text && args.is_full_text === "true") || (args.is_custom_query && args.is_custom_query === "true")) {
				query = args.query;
			} else {
				if (args.filters) {
					query = await this.makeQueryFromFilter(args.filters);
				} else {
					if (args.query) {
						query = args.query + " AND (deleted = false OR deleted IS MISSING) ";
					} else {
						query = "(deleted = false OR deleted IS MISSING) ";
					}
				}
			}

			if (args.limit) {
				args.limit = Number(args.limit)
			}
			if (args.offset) {
				args.offset = Number(args.offset)
			}

			if (!args.sort) {
				args.sort = "created_at";
			}

			await this.run_hook(this, 'get', 'before');
			let results
			if (args.is_custom_query && args.is_custom_query === "true") {
				//console.log("custom",query)
				results = await this.database.query(query);
				await this.run_hook(results, 'get', 'after');
				await this.do_serialize(results, 'read', {}, args);
				//console.log("1x")
				return results;
			} else {
				if (args.is_full_text && args.is_full_text === "true") {
					results = await this.database.full_text_search(this.type, query || "", args.limit, args.offset);
				} else {
					try {
						results = await this.database.search(this.type, args.columns || "", query || "", args.limit, args.offset, args.sort, args.do_count, args.statement_consistent);
					} catch (e) {
						console.log(e.stack)
					}
				}
				try {
					await this.run_hook(results.data, 'get', 'after');
					await this.do_serialize(results.data, 'read', {}, args);
				} catch (e) {
					console.log(e);
				}
				//console.log(results, results.total != null, results.total != undefined);
				if (results.total != null || results.total != undefined) {
					return results;
				}
				return results.data;
			}


		} catch (e) {
			console.log(e.stack);
			throw e;
		}

	}

	async run_hook(data, op, when, old_data) {
		try {
			if (this.hooks && this.hooks()[op] && this.hooks()[op][when]) {
				if (_.isArray(this.hooks()[op][when])) {
					for (var i of this.hooks()[op][when]) {
						data = await i(data, old_data);
					}
				} else {
					data = await this.hooks()[op][when](data, old_data);
				}
			}
			return data;
		} catch (e) {
			throw e;
		}
	}

	/*async run_hook(data, obj, context){
		try{


			if(obj) {
				if (_.isArray(obj)) {
					for (var i of obj) {
						await i.apply(this, [data, context]);
					}
				} else if (_.isFunction(obj)) {
					await obj.apply(this, [data, context]);
				}
			}
		}catch (e){
			console.log(e.stack);
		}
	}*/

	defaultCleaners() {
		return [
			async (data) => {
				try {
					if (_.isArray(data)) {
						for (let i of data) {
							delete i['deleted'];
							delete i['type'];
						}
					} else {
						delete data['deleted'];
						delete data['type'];
					}
					return data;
				} catch (e) {
					throw e;
				}
			}
		]
	}

	defaultModifiers() {
		return [
			async (data) => {
				try {
					return data;
				} catch (e) {
					throw e;
				}
			}
		]
	}

	serializer() {
		return {
			read: {
				modifiers: [
					/*async (data)=>{
						try{
							if(_.isArray(data)){
								//modifiers for arrays

							}else{

								//modifiers for single
							}
							return data;
						}catch(e){
							throw new Error(e);
						}
					}*/
				],
				cleaners: [
					//this is the default cleaner is none is specifier. add cleaner to override the default
					/*async (data)=>{
						try{
							if(_.isArray(data)){
								for(let i of data){
									delete i['deleted'];
									delete i['type'];
								}
							}else{
								delete data['deleted'];
								delete data['type'];
							}
							return data;
						}catch(e){
							throw new Error(e);
						}
					}*/
				]
			},
			write: {
				modifiers: [
					async (data) => {
						try {
							return data;
						} catch (e) {
							throw e;
						}
					}
				],
				cleaners: [
					async (data) => {
						try {
							return data;
						} catch (e) {
							throw e;
						}
					}
				]
			}
		}
	}

	hooks() {
		return {
			create: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}

				}
			},
			get: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			},
			update: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			},
			delete: {
				before: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				},
				after: async (data) => {
					try {
						//your code comes here
						return data;
					} catch (e) {
						throw e;
					}
				}
			}
		}
	}

	shouldSerializerRun(args, type){
		//console.log(args);
		if(args){
			if (args.skip_serialize == true){
				return false
			}
			if(type == 'read'){
				if (args.skip_read_serialize == true){
					return false;
				}
			} else if (type == 'write'){
				if (args.skip_write_serialize == true) {
					return false;
				}
			}
		}
		return true;
	}

	async do_serialize(data, type, old_data, args) {
		try {
			if (this.shouldSerializerRun(args, type)) {
				
						let serializer_obj = this.serializer();
						let active_modifiers = serializer_obj[type]['modifiers']

						if (!active_modifiers) {
							await this.defaultModifiers();
						}

						if (active_modifiers) {
							if (_.isArray(active_modifiers)) {
								for (let i of active_modifiers) {
									try {
										data = await i(data, old_data);
									} catch (e) {
										console.log(e.stack);
									}
								}
							} else {
								data = await active_modifiers(data, old_data);
							}
						}
						//apply cleaners
						let active_cleaners = serializer_obj[type]['cleaners']
						if (type == 'read') {
							if (!active_cleaners) {
								active_cleaners = await this.defaultCleaners();
							} else {
								if (_.isArray(active_cleaners)) {
									if (active_cleaners.length == 0) {
										active_cleaners = await this.defaultCleaners();
									}
								}
							}
						}

						if (active_cleaners) {
							if (_.isArray(active_cleaners)) {
								for (let i of active_cleaners) {
									data = await i(data);
								}
							} else {
								data = await active_cleaners(data);
							}
						}

						//get defaults from peroperties
						let defaults = {}
						if (this.props) {
							for (let i in this.props) {
								if (this.props[i].defaults != undefined && this.props[i].defaults[type] != undefined) {
									defaults[i] = this.props[i].defaults[type];
								}
							}
						}

						//apply defaults
						if (_.isArray(data)) {
							for (let i of data) {
								_.defaults(i, defaults);
							}
						} else {
							_.defaults(data, defaults);
						}

						return data;
					}

		} catch (e) {
			console.log(e.stack);
			throw e;
		}
	}
}
